import http.server
import socketserver
import subprocess
import os
import signal
import urllib.parse
import re
import urllib.request
import zipfile
import stat

PORT = 8001
HTML_TEMPLATE_PATH = os.path.join(os.path.dirname(__file__), 'taskmgr.html')
CLI_APPS_DIR = os.path.expanduser('~/usr/local/bin')  # Installation directory for CLI apps
WEB_APPS_DIR = os.path.expanduser('~/apps')          # Installation directory for Web apps
AVAILABLE_APPS_URL = 'http://berrystore.sw7ft.com/bins/'  # URL to fetch Command Line Utilities .zip files
WEB_APPS_URL = 'http://berrystore.sw7ft.com/apps/'      # URL to fetch Web Apps
app_ports = {}  # Store the mapping of PID to port

class TaskManagerHandler(http.server.BaseHTTPRequestHandler):

    def do_GET(self):
        parsed_path = urllib.parse.urlparse(self.path)
        if self.path.startswith('/action'):
            # Process action (start/stop/install/delete)
            params = urllib.parse.parse_qs(parsed_path.query)
            action = params.get('action', [''])[0]
            app_name = params.get('app_name', [''])[0]
            pid = params.get('pid', [''])[0]
            app_type = params.get('app_type', ['cli'])[0]  # Default to 'cli' if not provided
            if action == 'start' and app_name:
                self.start_app(app_name, app_type)
            elif action == 'stop' and pid:
                self.stop_app(pid)
            elif action == 'install' and app_name:
                self.install_app(app_name, app_type)
            elif action == 'delete' and app_name:
                self.delete_app(app_name, app_type)
            # Redirect back to home page
            self.send_response(303)
            self.send_header('Location', '/')
            self.end_headers()
        else:
            # Serve the main page using the HTML template
            self.send_response(200)
            self.send_header('Content-type', 'text/html')
            self.end_headers()
            html = self.generate_html()
            self.wfile.write(html.encode())

    def install_app(self, app_name, app_type='cli'):
        try:
            if app_type == 'cli':
                download_url = urllib.parse.urljoin(AVAILABLE_APPS_URL, app_name)
                install_dir = CLI_APPS_DIR
            elif app_type == 'web':
                download_url = urllib.parse.urljoin(WEB_APPS_URL, app_name)
                install_dir = WEB_APPS_DIR
            else:
                print(f"Unknown app type: {app_type}")
                return

            print(f"Downloading {download_url}")

            with urllib.request.urlopen(download_url) as response:
                if response.status != 200:
                    print(f"Failed to download {download_url}: Status {response.status}")
                    return
                zip_data = response.read()

            if not os.path.exists(install_dir):
                os.makedirs(install_dir)
                print(f"Created installation directory at: {install_dir}")

            zip_path = os.path.join(install_dir, app_name)
            with open(zip_path, 'wb') as f:
                f.write(zip_data)
            print(f"Downloaded {app_name} to {zip_path}")

            with zipfile.ZipFile(zip_path, 'r') as zip_ref:
                if app_type == 'web':
                    # Extract to a subdirectory named after the app (without .zip)
                    app_folder = os.path.splitext(app_name)[0]
                    extract_path = os.path.join(install_dir, app_folder)
                    os.makedirs(extract_path, exist_ok=True)
                    zip_ref.extractall(extract_path)
                    print(f"Extracted {app_name} to {extract_path}")
                else:
                    # For CLI apps, extract directly to install_dir
                    zip_ref.extractall(install_dir)
                    print(f"Extracted {app_name} to {install_dir}")

            os.remove(zip_path)
            print(f"Removed zip file: {zip_path}")

            for root, dirs, files in os.walk(install_dir):
                for file in files:
                    file_path = os.path.join(root, file)
                    if os.path.isfile(file_path):
                        st = os.stat(file_path)
                        os.chmod(file_path, st.st_mode | stat.S_IEXEC)
                        print(f"Set executable permissions for: {file_path}")

            current_path = os.environ.get('PATH', '')
            if install_dir not in current_path.split(os.pathsep):
                print(f"Note: {install_dir} is not in your PATH.")
                print("Please add the following line to your ~/.profile to include it in your PATH:")
                print(f'export PATH="{install_dir}:$PATH"')

        except Exception as e:
            print(f"Error installing {app_name}: {e}")

    def delete_app(self, app_name, app_type='cli'):
        try:
            if app_type == 'web':
                app_folder = os.path.splitext(app_name)[0]
                app_path = os.path.join(WEB_APPS_DIR, app_folder)
            else:
                app_path = os.path.join(CLI_APPS_DIR, app_name)

            if not os.path.exists(app_path):
                print(f"App {app_name} does not exist in the installation directory")
                return

            if os.path.isfile(app_path) or os.path.islink(app_path):
                os.remove(app_path)
                print(f"Deleted app: {app_path}")
            elif os.path.isdir(app_path):
                subprocess.run(['rm', '-rf', app_path])
                print(f"Deleted directory app: {app_path}")

        except Exception as e:
            print(f"Error deleting {app_name}: {e}")

    def start_app(self, app_name, app_type='cli'):
        try:
            if app_type == 'web':
                app_folder = os.path.splitext(app_name)[0]
                app_path = os.path.join(WEB_APPS_DIR, app_folder, 'app.py')
            else:
                app_path = os.path.join(CLI_APPS_DIR, app_name)

            if not os.path.exists(app_path):
                print(f"App {app_name} does not exist in the installation directory")
                return

            print(f"Attempting to start app: {app_path}")

            python_path = os.path.expanduser('~/python3_ntoarmv7-qnx-static/tools/python3/python3')

            if not os.path.exists(python_path):
                print(f"Python interpreter not found at {python_path}, falling back to 'python3'")
                python_path = 'python3'

            command = [python_path, '-u', app_path]

            print(f"Running command: {' '.join(command)}")

            process = subprocess.Popen(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)

            while True:
                output_line = process.stdout.readline()
                if not output_line:
                    break
                output_line = output_line.strip()
                print(f"App output: {output_line}")

                match = re.search(r'(\d{4,5})', output_line)
                if match:
                    port = match.group(1)
                    app_ports[process.pid] = port
                    print(f"App started on port {port}")
                    break

            print(f'Successfully started app {app_name} with PID {process.pid}')
        except Exception as e:
            print(f'Error starting {app_name}: {e}')

    def stop_app(self, pid):
        try:
            os.kill(int(pid), signal.SIGTERM)
            print(f'Stopped process {pid}')
            if int(pid) in app_ports:
                del app_ports[int(pid)]
        except Exception as e:
            print(f'Error stopping process {pid}: {e}')

    def generate_html(self):
        try:
            with open(HTML_TEMPLATE_PATH, 'r') as file:
                html = file.read()

            # Insert running processes
            processes_html = self.generate_processes_html()
            html = html.replace('<!-- running_processes -->', processes_html)

            # Insert installed apps
            installed_apps_html = self.generate_installed_apps_html()
            html = html.replace('<!-- installed_apps -->', installed_apps_html)

            # Insert available command line utilities
            available_cli_html = self.generate_available_cli_html()
            html = html.replace('<!-- available_cli_apps -->', available_cli_html)

            # Insert available web apps
            available_web_html = self.generate_available_web_apps_html()
            html = html.replace('<!-- available_web_apps -->', available_web_html)

            return html
        except Exception as e:
            print(f'Error loading HTML template: {e}')
            return '<html><body><h1>Error loading HTML template</h1></body></html>'

    def generate_processes_html(self):
        processes = self.get_python_processes()
        rows = ''
        for pid, cmd in processes:
            launch_link = ''
            if int(pid) in app_ports:
                port = app_ports[int(pid)]
                launch_link = f' - <a href="http://127.0.0.1:{port}" target="_blank">Launch UI</a>'
            rows += f'<tr><td>{pid}</td><td class="scrollable-command">{cmd}</td><td><a href="/action?action=stop&pid={pid}">Stop</a>{launch_link}</td></tr>'
        return rows

    def generate_installed_apps_html(self):
        cli_apps = self.scan_apps_directory(CLI_APPS_DIR, 'cli')
        web_apps = self.scan_apps_directory(WEB_APPS_DIR, 'web')
        all_apps = cli_apps + web_apps

        if not all_apps:
            return '<p>No installed apps found.</p>'

        apps_html = '<table>'
        apps_html += '<tr><th>App Name</th><th>Type</th><th>Action</th></tr>'
        for app, app_type in all_apps:
            app_encoded = urllib.parse.quote(app)
            if app_type == 'web':
                apps_html += f'<tr><td>{app}</td><td>Web App</td><td><a href="/action?action=start&app_name={app_encoded}&app_type=web">Start</a> | <a href="/action?action=delete&app_name={app_encoded}&app_type=web">Delete</a></td></tr>'
            else:
                apps_html += f'<tr><td>{app}</td><td>CLI</td><td><a href="/action?action=delete&app_name={app_encoded}&app_type=cli">Delete</a></td></tr>'
        apps_html += '</table>'
        return apps_html

    def generate_available_cli_html(self):
        try:
            available_zips = self.fetch_available_zips()
            installed_cli_apps = set(app for app, _ in self.scan_apps_directory(CLI_APPS_DIR, 'cli'))
            # Exclude installed apps
            available_zips = [z for z in available_zips if os.path.splitext(z)[0] not in installed_cli_apps]

            if not available_zips:
                return '<p>No Command Line Utilities available.</p>'

            apps_html = '<div class="apps-container">'
            for zip_file in available_zips:
                app_name = os.path.splitext(zip_file)[0]
                app_encoded = urllib.parse.quote(zip_file)
                apps_html += f'''
                <div class="app-item">
                    <div class="app-icon"></div>
                    <div class="app-name">{app_name}</div>
                    <a href="/action?action=install&app_name={app_encoded}&app_type=cli" class="install-button">Install</a>
                </div>
                '''
            apps_html += '</div>'
            return apps_html
        except Exception as e:
            print(f"Error generating available CLI apps HTML: {e}")
            return '<p>Error loading Command Line Utilities.</p>'

    def generate_available_web_apps_html(self):
        try:
            available_zips = self.fetch_web_apps()
            installed_web_apps = set(app for app, _ in self.scan_apps_directory(WEB_APPS_DIR, 'web'))
            # Exclude installed apps
            available_zips = [z for z in available_zips if os.path.splitext(z)[0] not in installed_web_apps]

            if not available_zips:
                return '<p>No Web Apps available.</p>'

            apps_html = '<div class="apps-container">'
            for zip_file in available_zips:
                app_name = os.path.splitext(zip_file)[0]
                app_encoded = urllib.parse.quote(zip_file)
                apps_html += f'''
                <div class="app-item">
                    <div class="app-icon"></div>
                    <div class="app-name">{app_name}</div>
                    <a href="/action?action=install&app_name={app_encoded}&app_type=web" class="install-button">Install</a>
                </div>
                '''
            apps_html += '</div>'
            return apps_html
        except Exception as e:
            print(f"Error generating available Web Apps HTML: {e}")
            return '<p>Error loading Web Apps.</p>'

    def fetch_available_zips(self):
        try:
            with urllib.request.urlopen(AVAILABLE_APPS_URL) as response:
                if response.status != 200:
                    print(f"Failed to fetch Command Line Utilities: Status {response.status}")
                    return []
                html_content = response.read().decode()

            zip_files = re.findall(r'href=["\']([^"\']+\.zip)["\']', html_content)
            zip_files = [os.path.basename(z) for z in zip_files]
            return zip_files
        except Exception as e:
            print(f"Error fetching Command Line Utilities zips: {e}")
            return []

    def fetch_web_apps(self):
        try:
            with urllib.request.urlopen(WEB_APPS_URL) as response:
                if response.status != 200:
                    print(f"Failed to fetch Web Apps: Status {response.status}")
                    return []
                html_content = response.read().decode()

            zip_files = re.findall(r'href=["\']([^"\']+\.zip)["\']', html_content)
            zip_files = [os.path.basename(z) for z in zip_files]
            return zip_files
        except Exception as e:
            print(f"Error fetching Web Apps zips: {e}")
            return []

    def get_python_processes(self):
        processes = []
        try:
            output = subprocess.check_output(['pidin', 'ar']).decode()
            lines = output.strip().split('\n')
            for line in lines:
                if 'python' in line:
                    parts = line.split()
                    pid = parts[0]
                    cmd = ' '.join(parts[1:])
                    if '/data/' in cmd:
                        index = cmd.find('/data/') + len('/data/')
                        cmd = cmd[index:]
                    processes.append((pid, cmd))
        except Exception as e:
            print(f'Error getting processes: {e}')
        return processes

    def scan_apps_directory(self, directory, app_type):
        apps = []
        try:
            if not os.path.exists(directory):
                return apps
            for root, dirs, files in os.walk(directory):
                for file in files:
                    if app_type == 'web':
                        if file == 'app.py':
                            app_folder = os.path.relpath(root, directory)
                            apps.append((app_folder, 'web'))
                    else:
                        if not file.endswith('.py'):  # Assuming CLI apps are executables without .py
                            rel_path = os.path.relpath(os.path.join(root, file), directory)
                            apps.append((rel_path, 'cli'))
            return apps
        except Exception as e:
            print(f'Error scanning apps directory: {e}')
            return apps

def check_path():
    # Check CLI_APPS_DIR
    current_path = os.environ.get('PATH', '')
    paths = current_path.split(os.pathsep)
    if CLI_APPS_DIR not in paths:
        print(f"Warning: {CLI_APPS_DIR} is not in your PATH.")
        print("Please add the following line to your ~/.profile to include it in your PATH:")
        print(f'export PATH="{CLI_APPS_DIR}:$PATH"')
    else:
        print(f"{CLI_APPS_DIR} is already in your PATH.")

    # Check WEB_APPS_DIR
    if WEB_APPS_DIR not in paths:
        print(f"Note: {WEB_APPS_DIR} is not in your PATH.")
        print("If you want to run web apps from the command line, consider adding it to your PATH:")
        print(f'export PATH="{WEB_APPS_DIR}:$PATH"')
    else:
        print(f"{WEB_APPS_DIR} is already in your PATH.")

if __name__ == '__main__':
    check_path()  # Check PATH before starting the server
    with socketserver.TCPServer(("", PORT), TaskManagerHandler) as httpd:
        print(f"Serving on port {PORT}")
        httpd.serve_forever()
